import os
import io
import unittest
import tempfile
import gzip
import json

from findopendata.storage.local import LocalStorage
from findopendata.parsers.avro import avro2json
from findopendata.parsers.jsonl import jsonl2json

test_obj = {
    "name" : "First name Last name",
    "email" : "example@email.com",
    "age" : 42,
    "accounts": [123, 432, 2123],
}

test_file_content = """
h1,h2,h3
a,b,c
e,f,g
1,2,3
"""

test_avro_records = [
    {"h1" : "a", "h2": "b", "h3": "c"},
    {"h1" : "d", "h2": "e", "h3": "f"},
    {"h1" : "1", "h2": "2", "h3": "3"},
]

test_avro_schema = {
    "name": "root",
    "type": "record",
    "fields": [
        {"name": "h1", "type": "string"}, 
        {"name": "h2", "type": "string"}, 
        {"name": "h3", "type": "string"}, 
    ],
}

class TestLocalStorage(unittest.TestCase):

    def test_put_and_get_object(self):
        with tempfile.TemporaryDirectory() as root:
            storage = LocalStorage(root)
            blob = storage.put_object(test_obj, "test_blob")
            self.assertEqual(blob.name, "test_blob")
            self.assertTrue(blob.size > 0)
            self.assertTrue(os.path.exists(os.path.join(root, "test_blob")))

            obj = storage.get_object("test_blob")
            self.assertEqual(test_obj, obj)
    
    def test_put_and_get_file(self):
        with tempfile.TemporaryDirectory() as root:
            storage = LocalStorage(root)
            fileobj = io.BytesIO(test_file_content.encode("utf-8"))
            blob = storage.put_file(fileobj, "test_blob")
            self.assertEqual(blob.name, "test_blob")
            self.assertGreater(blob.size, 0)
            self.assertTrue(os.path.exists(os.path.join(root, "test_blob")))

            with storage.get_file("test_blob") as f:
                data = f.read().decode("utf-8")
                self.assertEqual(data, test_file_content)
            
    def test_put_avro(self):
        with tempfile.TemporaryDirectory() as root:
            storage = LocalStorage(root)
            blob = storage.put_avro(test_avro_schema, test_avro_records, 
                    "test_blob")
            self.assertEqual(blob.name, "test_blob")
            self.assertGreater(blob.size, 0)
            self.assertTrue(os.path.exists(os.path.join(root, "test_blob")))

            with storage.get_file("test_blob") as f:
                records = avro2json(f)
                for r1, r2 in zip(records, test_avro_records):
                    self.assertEqual(dict(r1), r2)
            
    def test_put_json(self):
        with tempfile.TemporaryDirectory() as root:
            storage = LocalStorage(root)
            blob = storage.put_json(test_avro_records, "test_json_blob")
            self.assertEqual(blob.name, "test_json_blob")
            self.assertGreater(blob.size, 0)
            self.assertTrue(os.path.exists(os.path.join(root, 
                    "test_json_blob")))

            with storage.get_file("test_json_blob") as f:
                f = io.TextIOWrapper(gzip.GzipFile(fileobj=f, mode="rb"))
                records = [json.loads(line) for line in f]
                for r1, r2 in zip(records, test_avro_records):
                    self.assertEqual(r1, r2)
            

if __name__ == "__main__":
    unittest.main()


import time

import requests
import dateutil.parser
import dateutil.tz


def read_api(api_url, start=0, page_size=50, retries=3,
        wait_between_retries=5):
    """Scrolls through the CKAN package_search API to obtain packages.

    Args:
        api_url: the CKAN API endpoint URL (i.e., https://data.gov.uk).
        start: the starting record index.
        page_size: the number of records per each request.
        retries: the number of retries when an error is encountered.
        wait_between_retries: the seconds to wait between retries.
    """
    url = api_url.rstrip("/") + "/api/3/action/package_search"
    sess = requests.session()
    while True:
        resp = sess.get(url, params={"start" : start,
                                     "rows"  : page_size})
        try:
            resp.raise_for_status()
        except Exception as e:
            if retries == 0:
                raise e
            retries -= 1
            time.sleep(wait_between_retries)
            continue
        results = resp.json()["result"]["results"]
        if len(results) == 0:
            break
        for package in results:
            yield package
        start += page_size


def parse_timestamp(timestamp_str):
    try:
        timestamp = dateutil.parser.parse(timestamp_str)
        # Set tzinfo to UTC when not available
        if timestamp.tzinfo is None:
            timestamp = timestamp.replace(tzinfo=dateutil.tz.tzutc())
    except ValueError:
        return None
    return timestamp


def extract_timestamp_from_package(package):
    package_modified = None
    if "modified" in package:
        package_modified = package["modified"]
    elif "metadata_modified" in package:
        package_modified = package["metadata_modified"]
    elif "metadata_created" in package:
        package_modified = package["metadata_created"]
    if package_modified is None:
        return None
    return parse_timestamp(package_modified)


def extract_timestamp_from_resource(resource):
    resource_modified = None
    if "created" in resource and resource["created"] is not None:
        resource_modified = resource["created"]
    if "revision_timestamp" in resource and \
            resource["revision_timestamp"] is not None:
        resource_modified = resource["revision_timestamp"]
    if "last_modified" in resource and resource["last_modified"] is not None:
        resource_modified = resource["last_modified"]
    if resource_modified is None:
        return None
    return parse_timestamp(resource_modified)



#!/usr/bin/env python
import argparse
import sys

from findopendata.ckan_crawler import add_ckan_apis
from findopendata.socrata_crawler import add_socrata_discovery_apis

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
            description="Start dataset harvesting tasks.")
    parser.add_argument("-u", "--force-update", action='store_true',
            help="Whether to force update a dataset even if the modified time "
                "is prior to the last updated time.")
    args = parser.parse_args(sys.argv[1:])

    # CKAN
    add_ckan_apis.delay(force_update=args.force_update)
    # Socrata
    add_socrata_discovery_apis.delay(force_update=args.force_update)



import csv
import io
import itertools
from collections import OrderedDict

from .encoding import guess_encoding_from_buffer, guess_encoding_from_stream


_sniffer = csv.Sniffer()


def _is_number(x):
    try:
        a = float(x)
    except ValueError:
        return False
    else:
        return True


def csv2json(fileobj_binary, 
        guess_encoding_bytes=8192, 
        guess_dialect_lines=5,
        guess_header_rows=10,
        allow_no_header=False, 
        header_prefix="Column-",
        min_header_count=2):
    """Read a CSV file and get an iterator of JSON records as Python dictionaries.

    Args:
        fileobj_binary: a binary file object that supports seek().
        guess_encoding_bytes: the number of bytes in the beginning of the file
            to be used to guess the text encoding. If sets to -1, all the file
            will potentially be read to determine the encoding, before any data
            is parsed.
        guess_dialect_lines: the number of text lines to read to guess the 
            CSV dialect.
        guess_header_rows: the number of lines from the beginning of the file
            to be used to guess the headers.
        allow_no_header: whether to raise ValueError if no header row was found.
        header_prefix: the string prefix used for assigning headers
            to CSV tables without a header row.
        min_header_count: the minimum number of headers to consider a valid
            CSV table.

    Returns: an iterator of JSON records as Python dictionaries.
    """

    if guess_encoding_bytes == -1:
        # Guess encoding by reading the file as a stream.
        encoding = guess_encoding_from_stream(fileobj_binary)
    else:
        # Guess encoding by reading the beginning of the file.
        head = fileobj_binary.read(guess_encoding_bytes)
        encoding = guess_encoding_from_buffer(head)
    
    # Rewind
    fileobj_binary.seek(0)
    
    # Read decoded text.
    fileobj = io.TextIOWrapper(fileobj_binary, encoding=encoding, newline='')

    # Guess dialect and headers
    dialect = _sniffer.sniff("".join([fileobj.readline() 
            for _ in range(guess_dialect_lines)]))

    # Rewind
    fileobj.seek(0)

    # Wrap the binary file with text file reader to create csv reader
    reader = csv.reader(fileobj, dialect)

    # Figure out the first row that looks like a header using the
    # first few rows.
    head = [row for _, row in zip(range(guess_header_rows), reader)]
    headers = None
    header_row_pos = 0
    for i, row in enumerate(head):
        if all(len(v.strip()) > 0 and not _is_number(v) for v in row):
            headers = row
            header_row_pos = i
            break
    # Assign default headers if none found
    if headers is None:
        if not allow_no_header:
            raise ValueError("No header row found.")
        ncol = max(len(row) for row in head)
        headers = ["{}{}".format(header_prefix, i) for i in range(ncol)]
        header_row_pos = -1
    if len(headers) < min_header_count:
        raise ValueError("Not enough header (min {}) to be valid".format(
                min_header_count))

    # Yield records.
    rows = itertools.chain(head[header_row_pos+1:], reader)
    for row in rows:
        yield OrderedDict(zip(headers, row))



from collections import OrderedDict

from .column_sketch import ColumnSketch


class TableSketch:
    """A TableSketch contains summaries of a table, including the column 
    sketches.

    Args:
        record_sample_size: the number of record to include in the sample.
        column_sketch_kwargs: keyword arguments for ColumnSketch's constructor.
    """

    def __init__(self, record_sample_size=20, **column_sketch_kwargs):
        self._column_sketches = {}
        self._record_sample_size = record_sample_size
        self._sample = []
        self._column_names = []
        self._column_sketch_kwargs = column_sketch_kwargs
    
    @property
    def column_sketches(self):
        """Column sketches in the order of column names."""
        return [self._column_sketches[name] for name in self._column_names]
    
    @property
    def record_sample(self):
        """Record samples as a list of dict."""
        return self._sample
    
    @property
    def column_names(self):
        """Column names in the order from left to right."""
        return self._column_names
    
    def update(self, record):
        # Check type.
        if not isinstance(record, OrderedDict):
            raise TypeError("record must be an OrderedDict")
        # Assign column names.
        if not self._column_names:
            self._column_names = list(record.keys())
        # Update column sketches.
        for column_name, value in record.items():
            if column_name not in self._column_sketches:
                self._column_sketches[column_name] = ColumnSketch(column_name, 
                        **self._column_sketch_kwargs)
            self._column_sketches[column_name].update(value)
        # Update record sample.
        if len(self._sample) < self._record_sample_size:
            self._sample.append(dict(record))



import io

import simplejson as json

from .encoding import guess_encoding_from_buffer


def jsonl2json(fileobj_binary, guess_encoding_bytes=8192):
    """Read a JSONL (newline-delimited JSON) file and get an iterator of
    JSON records as Python dictionaries.

    Args:
        fileobj_binary: a binary file object that supports seek().
        guess_encoding_bytes: the number of bytes in the beginning of the file
            to be used to guess the text encoding.

    Returns: an iterator of JSON records as Python dictionaries.
    """

    # Guess encoding
    head = fileobj_binary.read(guess_encoding_bytes)
    encoding = guess_encoding_from_buffer(head)

    # Rewind
    fileobj_binary.seek(0)

    # Wrap the binary file with text file reader to create line reader
    fileobj = io.TextIOWrapper(fileobj_binary, encoding=encoding, newline='')

    # Return records
    for line in fileobj:
        yield json.loads(line)



import io
import base64
import contextlib
import urllib
import collections
import gzip

import fastavro
import simplejson as json
from azure.storage.blob import BlobBlock, ContainerClient, BlobClient

from .base import Blob, BlobStorage


def _encode_base64(data):
    if isinstance(data, str):
        data = data.encode("utf-8")
    encoded = base64.b64encode(data)
    return encoded.decode("utf-8")


def _block_id(offset):
    return urllib.parse.quote(_encode_base64("{0:032d}".format(offset)))


class AzureBlobReader(io.BufferedIOBase):
    def __init__(
        self,
        blob_client: BlobClient,
    ):
        self._blob_client = blob_client
        self._start = 0
        self._buf = bytearray()
        self._chunks = None

    def writable(self):
        return False

    def read(self, size=-1):
        if self._start == 0:
            self._chunks = self._blob_client.download_blob().chunks()
        while (not size or size < 0) or len(self._buf) < size:
            chunk = next(self._chunks, None)
            if chunk is None:
                break
            self._buf += chunk
            self._start += len(chunk)
        data = self._buf[:size]
        self._buf = self._buf[size:]
        return bytes(data)

    def seek(self, offset):
        if offset != 0:
            raise ValueError("Only seeking to beginning is supported.")
        self._start = offset
        self._buf.clear()
        self._chunks = None


class AzureBlobWriter(io.BufferedIOBase):
    def __init__(
        self,
        blob_client: BlobClient,
        block_size: int = 4 * 1024 * 1024,
    ):
        self._blob_client = blob_client
        self._block_size = block_size
        self._buf = bytearray()
        self._block_offset = 0
        self._blocks = collections.deque([])
        self._closed = False
        self._position = 0

    def readable(self):
        return False

    def write(self, b):
        self._buf += bytes(b)
        while len(self._buf) >= self._block_size:
            block_id = _block_id(self._block_offset)
            self._blob_client.stage_block(
                block_id=block_id, data=bytes(self._buf[: self._block_size])
            )
            self._blocks.append(BlobBlock(block_id))
            self._block_offset += 1
            self._buf = self._buf[self._block_size :]
        self._position += len(b)
        return len(b)

    def flush(self):
        if self._buf:
            block_id = _block_id(self._block_offset)
            self._blob_client.stage_block(block_id=block_id, data=bytes(self._buf))
            self._blocks.append(BlobBlock(block_id))
            self._block_offset += 1
            self._buf.clear()

    def close(self):
        if self._closed:
            return
        self.flush()
        # Put the block list to make the blob.
        self._blob_client.commit_block_list(list(self._blocks))
        self._closed = True
        self._blocks.clear()

    def closed(self):
        return self._closed

    def tell(self):
        return self._position

    def seekable(self):
        return False


class AzureStorage(BlobStorage):
    """Azure storage provider that utilizes the Azure blob storage.

    Args:
        connection_string: See http://azure.microsoft.com/en-us/documentation/articles/storage-configure-connection-string/
            for the connection string format.
        container_name: the name of the blob container in which all blobs
            are stored.

    """

    def __init__(
        self, connection_string, container_name, block_size: int = 4 * 1024 * 1024
    ):
        self._container = ContainerClient.from_connection_string(
            connection_string, container_name
        )
        if not self._container.exists():
            raise ValueError("Container does not exist: " + container_name)
        self._container_name = container_name
        self._block_size = block_size

    def get_object(self, blob_name):
        blob_client = self._container.get_blob_client(blob_name)
        blob_content = blob_client.download_blob().content_as_text()
        return json.loads(blob_content)

    def put_object(self, obj, blob_name):
        blob_client = self._container.get_blob_client(blob_name)
        blob_content = json.dumps(obj).encode("utf-8")
        blob_client.upload_blob(blob_content, overwrite=True)
        return Blob(blob_name, len(blob_content))

    @contextlib.contextmanager
    def get_file(self, blob_name):
        try:
            stream = AzureBlobReader(self._container.get_blob_client(blob_name))
            yield stream
        finally:
            stream.close()

    def put_file(self, fileobj, blob_name):
        self._container.upload_blob(blob_name, fileobj, overwrite=True)
        size = fileobj.tell()
        return Blob(blob_name, size)

    def put_avro(self, schema, records, blob_name, codec="snappy"):
        writer = AzureBlobWriter(
            self._container.get_blob_client(blob_name), block_size=self._block_size
        )
        fastavro.writer(writer, schema, records, codec)
        writer.close()
        size = writer.tell()
        return Blob(blob_name, size)

    def put_json(self, records, blob_name, gzip_compress=True):
        writer = AzureBlobWriter(
            self._container.get_blob_client(blob_name), block_size=self._block_size
        )
        newline = "\n"
        if gzip_compress:
            with gzip.open(writer, "wt") as f:
                for record in records:
                    f.write(json.dumps(record))
                    f.write(newline)
        else:
            with io.TextIOWrapper(writer) as f:
                for record in records:
                    f.write(json.dumps(record))
                    f.write(newline)
        writer.close()
        size = writer.tell()
        return Blob(blob_name, size)



from collections import OrderedDict

import fastavro
from genson import SchemaBuilder


JSON_TO_AVRO_TYPES = {
        "null" : "null",
        "boolean" : "boolean",
        "integer" : "int",
        "number" : "double",
        "string" : "string",
        "object" : "record",
        "array" : "array",
        }


def _avro_type(json_type):
    if isinstance(json_type, list):
        return [_avro_type(jt) for jt in json_type]
    return JSON_TO_AVRO_TYPES.get(json_type)


def _json_to_avro_schema(json_schema, field_names=None, name="Root"):
    if json_schema["type"] == "object":
        if "properties" not in json_schema:
            return {"type" : "record", "name": name}
        fields = []
        for prop in json_schema["properties"]:
            avro_type = _json_to_avro_schema(json_schema["properties"][prop],
                    name="{}_type".format(prop))
            if isinstance(avro_type, list):
                avro_type = ["null",] + avro_type
            else:
                avro_type = ["null", avro_type]
            fields.append({"name": prop, "type": avro_type})
        if field_names is not None:
            field_order = dict((field, i)
                    for i, field in enumerate(field_names))
            # In some cases field_names may not match the actual fields in the 
            # records -- put those fields at the end.
            fields = sorted(fields, 
                    key=lambda f: field_order.get(f["name"], len(field_order)))
        return {"type": "record", "name": name, "fields": fields}
    elif json_schema["type"] == "array":
        return {
                "type" : "array",
                "items": _json_to_avro_schema(
                    json_schema["items"],
                    name="{}_item_type".format(name),
                    ),
                }
    return _avro_type(json_schema["type"])


def _infer_schema(records, field_names=None):
    builder = SchemaBuilder()
    count = 0
    for record in records:
        builder.add_object(record)
        count += 1
    if count == 0:
        return {
                "type": "record",
                "name" : "Root",
                }
    #print(builder.to_json(indent=2))
    schema = _json_to_avro_schema(builder.to_schema(), field_names)
    #print(json.dumps(schema, indent=True))
    return schema


def _is_iterator(obj):
    if hasattr(obj, '__iter__') and \
            hasattr(obj, '__next__') and \
            callable(obj.__iter__) and \
            obj.__iter__() is obj:
        return True
    else:
        return False


class JSON2AvroRecords(object):
    """A wrapper reader class for reading JSON data (deserialized as Python
    dict objects) with schema inference."""

    def __init__(self, json_records, field_names=None, head=25000):
        """Initializes from reading JSON records with schema
        inference.

        Args:
            json_records: an iterator of JSON records (deserialized as Python
                dict objects).
            field_names: a list of field names used to order the Avro fields,
                optional.
            head: the number of records in the beginning to use for schema
                inference.
        """
        if not _is_iterator(json_records):
            raise ValueError("json_records must be an iterator.")
        self._json_records = json_records
        self._head = [r for _, r in zip(range(head), self._json_records)]
        self._schema = _infer_schema(self._head, field_names)

    @property
    def schema(self):
        return self._schema

    def get(self):
        for record in self._head:
            yield record
        for record in self._json_records:
            yield record


def avro2json(fileobj_binary):
    reader = fastavro.reader(fileobj_binary)
    schema = reader.writer_schema
    field_names = [f["name"] for f in schema["fields"]]
    for record in reader:
        yield OrderedDict((f, record[f]) for f in field_names)




import os
import time

import requests
import urllib3
import psycopg2
from psycopg2.extras import Json, RealDictCursor
from celery.utils.log import get_task_logger

from .celery import app
from .ckan import read_api, extract_timestamp_from_package, \
    extract_timestamp_from_resource
from .storage.objects import storage
from .download import download_to_local
from .util import temporary_directory, get_safe_filename
from .settings import crawler_configs, db_configs, gcp_configs
from .parsers.csv import csv2json
from .parsers.avro import JSON2AvroRecords


# Logger for tasks.
logger = get_task_logger(__name__)

# Dataset formats for which we have available parsers.
accepted_resource_formats = ["csv",]


@app.task(ignore_result=True)
def add_ckan_package(package, endpoint, blob_prefix):
    """Registers the CKAN package and starts tasks for retrieving and adding
    associated resources (i.e. files).

    Args:
        package: the CKAN package JSON.
        endpoint: the CKAN endpoint (i.e., data.gov.uk) without http://.
        blob_prefix: the prefix for all the blobs uploaded from this
            function, relative to the root.
    """
    package_id = package["id"]

    # Upload the package JSON.
    package_blob = storage.put_object(package,
            os.path.join(blob_prefix, endpoint, package_id, "package.json"))
    logger.info("(endpoint={} package={}) Saved package JSON.".format(endpoint,
        package_id))

    # Initialize Postgres connection.
    conn = psycopg2.connect(**db_configs)
    cur = conn.cursor()

    # Register the package.
    cur.execute("INSERT INTO findopendata.ckan_packages "
            "(endpoint, package_id, package_blob) "
            "VALUES (%s, %s, %s) "
            "ON CONFLICT (endpoint, package_id) DO UPDATE "
            "SET updated = current_timestamp, "
            "package_blob = EXCLUDED.package_blob RETURNING key;",
            (endpoint, package_id, package_blob.name))
    row = cur.fetchone()
    if row is None:
        raise RuntimeError("(endpoint={} package={}) Failed to fetch key".format(
            endpoint, package_id))
    package_key = row[0]
    logger.info("(endpoint={} package={}) Registered package.".format(endpoint,
        package_id))

    # Commit and close database connection.
    conn.commit()
    cur.close()
    conn.close()

    # Skip processing resources if none exists.
    if "resources" not in package:
        return

    resource_blob_prefix = os.path.join(blob_prefix, endpoint, package_id)
    # Start tasks for processing resources.
    for resource in package["resources"]:
        if "format" not in resource or resource["format"].strip().lower() \
                not in accepted_resource_formats:
            add_ckan_resource_no_download(package_key=package_key,
                    resource=resource)
        else:
            add_ckan_resource.delay(package_key=package_key, resource=resource,
                    package_path=resource_blob_prefix)


@app.task(ignore_result=True)
def add_ckan_resource_no_download(package_key, resource):
    """Register this CKAN resource without downloading the associated files.

    Args:
        package_key: the key of the package associated with the resource.
        resource: the portion of the JSON data in the CKAN package
            corresponds to the resource.
    """
    # Get attributes of this resource.
    resource_id = resource.get("id", None)
    if resource_id is None:
        raise ValueError("(package={}) no resource id found.".format(
            package_key))
    original_url = resource.get("url", None)
    if original_url is None:
        logger.warning("(package={}, resource={}) no url found".format(
            package_key, resource_id))
        return
    filename = None

    # Initialize Postgres connection
    conn = psycopg2.connect(**db_configs)
    cur = conn.cursor()

    # Save this resource.
    cur.execute("INSERT INTO findopendata.ckan_resources "
            "(package_key, resource_id, filename, original_url, raw_metadata) "
            "VALUES (%s, %s, %s, %s, %s) "
            "ON CONFLICT (package_key, resource_id) "
            "DO UPDATE "
            "SET updated = current_timestamp, "
            "filename = EXCLUDED.filename, "
            "original_url = EXCLUDED.original_url, "
            "raw_metadata = EXCLUDED.raw_metadata;",
            (package_key, resource_id, filename, original_url, Json(resource)))
    
    # Save and close database connection
    conn.commit()
    cur.close()
    conn.close()

    # Done
    logger.info("(package={}, resource={}, filename={}) Registered "
            "resource.".format(package_key, resource_id, filename))


@app.task(ignore_result=True)
def add_ckan_resource(package_key, resource, package_path):
    """Retrieves and adds CKAN resource files for the given resource.

    Args:
        package_key: the key of the package associated with the resource.
        resource: the portion of the JSON data in the CKAN package
            corresponds to the resource.
        package_path: the blob path to the directory corresponding
            to the package associated with the resource.
    """
    resource_id = resource.get("id", None)
    if resource_id is None:
        raise ValueError("(package={}) no resource id found.".format(
            package_key))
    original_url = resource.get("url", None)
    if original_url is None:
        logger.warning("(package={}, resource={}) no url found".format(
            package_key, resource_id))
        return

    # Initialize Postgres connection for checking resource.
    conn = psycopg2.connect(**db_configs)
    cur = conn.cursor()

    # Check if the same version of this resource has been processed.
    cur.execute("SELECT updated::timestamptz FROM findopendata.ckan_resources "
            "WHERE package_key = %s AND resource_id = %s;",
            (package_key, resource_id))
    row = cur.fetchone()
    if row is not None:
        last_registered = row[0]
        last_updated = extract_timestamp_from_resource(resource)
        if last_updated is not None and last_updated <= last_registered:
            logger.info("(package={}, resource={}) Skipping (updated: {} "
                    "registered: {})".format(package_key, resource_id,
                        last_updated, last_registered))
            return

    # Close database connection for checking resource.
    cur.close()
    conn.close()

    # Download and upload the resource.
    logger.info("(package={} resource={}) Saving resource from {}".format(
        package_key, resource_id, original_url))
    working_dir = crawler_configs.get("working_dir", "/tmp")
    with temporary_directory(working_dir) as parent_dir:

        # Download this resource.
        try:
            filename = download_to_local(original_url, parent_dir)
        except Exception as e:
            logger.warning("(package={} resource={}) Failed to download {}: "
                    "{}".format(package_key, resource_id, original_url, e))
            return

        # Save and register the resource file.
        # Build blob name
        blob_name = os.path.join(package_path, resource_id, filename)
        try:
            with open(os.path.join(parent_dir, filename), "rb") as f:
                resource_blob = storage.put_file(f, blob_name)
        except Exception as e:
            logger.warning("(package={} resource={}) Failed to save local "
                    "file {} to {}: {}".format(package_key, resource_id,
                        filename, blob_name, e))
            return
        logger.info("(package={} resource={} filename={}) Saved resource "
                "from {} to {}".format(package_key, resource_id, filename,
                    original_url, blob_name))

    # Initialize Postgres connection for registering resource.
    # A new connection is created here to prevent the download
    # from hogging the connection pool.
    conn = psycopg2.connect(**db_configs)
    cur = conn.cursor()

    # Register this resource.
    cur.execute("INSERT INTO findopendata.ckan_resources "
            "(package_key, resource_id, filename, resource_blob, "
            "original_url, file_size, raw_metadata) "
            "VALUES (%s, %s, %s, %s, %s, %s, %s) "
            "ON CONFLICT (package_key, resource_id) "
            "DO UPDATE "
            "SET updated = current_timestamp, "
            "resource_blob = EXCLUDED.resource_blob, "
            "original_url = EXCLUDED.original_url, "
            "file_size = EXCLUDED.file_size, "
            "raw_metadata = EXCLUDED.raw_metadata;",
            (package_key, resource_id, filename,
                resource_blob.name, original_url,
                resource_blob.size, Json(resource)))
    conn.commit()

    # Close database connection for registering resource.
    cur.close()
    conn.close()

    logger.info("(package={}, resource={}, filename={}) Registered "
            "resource.".format(package_key, resource_id, filename))


@app.task(ignore_result=True)
def add_ckan_packages_from_api(api_url, endpoint, blob_prefix, force_update):
    """Scrolls through the CKAN package_search API to obtain packages and
    starts tasks to retrieve and add those packages.

    Args:
        api_url: the CKAN API endpoint URL (i.e., https://data.gov.uk).
        endpoint: the CKAN API endpoint without scheme (i.e., data.gov.uk).
        blob_prefix: the prefix for all the blobs uploaded from this
            function, relative to the root.
        force_update: whether to force update packages in the registry. By
            default, packages with updated time before the previously
            registered time will be skipped.
    """
    if not force_update:
        logger.info("Reading last updated timestamps for endpoint {}".format(
                endpoint))
        conn = psycopg2.connect(**db_configs)
        cur = conn.cursor()
        cur.execute(r"""SELECT package_id, updated::timestamptz
                        FROM findopendata.ckan_packages
                        WHERE endpoint = %s""", (endpoint,))
        updated_times = dict((package_id, updated) 
                for package_id, updated in cur)
        cur.close()
        conn.close()

    logger.info("Reading CKAN API: {}".format(api_url))
    packages = read_api(api_url)
    for package in packages:
        if not force_update:
            metadata_modified = extract_timestamp_from_package(package)
            package_updated = updated_times.get(package["id"])
            if metadata_modified is not None and package_updated is not None \
                    and metadata_modified <= package_updated:
                continue
        add_ckan_package.delay(package, endpoint=endpoint, 
                blob_prefix=blob_prefix)


@app.task(ignore_result=True)
def add_ckan_apis(force_update):
    """Add CKAN API endpoints to the crawler."""
    conn = psycopg2.connect(**db_configs)
    cur = conn.cursor()
    cur.execute(r"""SELECT scheme, endpoint 
                    FROM findopendata.ckan_apis 
                    WHERE enabled = true""")
    def _get_api_url(scheme, endpoint):
        endpoint = endpoint.rstrip("/")
        return ("{}://{}".format(scheme, endpoint), endpoint)
    api_urls = [_get_api_url(scheme, endpoint) for scheme, endpoint in cur]
    cur.close()
    conn.close()
    for api_url, endpoint in api_urls:
        add_ckan_packages_from_api.delay(api_url=api_url,
                endpoint=endpoint,
                blob_prefix=crawler_configs.get("ckan_blob_prefix"),
                force_update=force_update)
        logger.info("Adding CKAN API {} to the crawler".format(api_url))



import os
import shutil

import rfc6266
import requests

from .util import get_safe_filename

def download_to_local(url, dir_name):
    """Downloads remote resource given its URL.

    Args:
        url: the URL to the resource.
        dir_name: the directory on the local filesystem to save the resource.

    Returns:
        filename: the filename (relative to the dir_name) of the downloaded
            resource. It may be different from the name of the remote resource
            because of sanitization.
    """

    # TODO: be able to verify SSL certificates from some publishers
    with requests.get(url, stream=True) as r:
        r.raise_for_status()

        # Guesst the proper filename to use
        filename = ""
        # 1. Try to use the content-disposition header if available
        if "content-disposition" in r.headers:
            filename = rfc6266.parse_headers(r.headers["content-disposition"],
                    relaxed=True).filename_unsafe
        # 2. Try to get it from the URL
        if filename == "":
            filename = url.rsplit("/", 1)[-1]
        # 3. Sanitize the filename, this handles empty filename right now
        filename = get_safe_filename(filename)

        # Download the file
        with open(os.path.join(dir_name, filename), "wb") as o:
            shutil.copyfileobj(r.raw, o)
    return filename



import logging

from ..settings import azure_configs
from ..settings import local_configs
from ..settings import gcp_configs
from .base import BlobStorage
from .gcp import GoogleCloudStorage
from .azure import AzureStorage
from .local import LocalStorage


def BlobStorageFactory(provider="local") -> BlobStorage:
    """Create a storage provider.

    Args:
        provider: the name of the storage provider. Choose among 
            `local` and `gcp`.
    
    Return: a storage provider of the class `BlobStorage`.
    """
    if provider == "local":
        root = local_configs.get("root")
        return LocalStorage(root)
    if provider == "gcp":
        project_id = gcp_configs.get("project_id")
        bucket_name = gcp_configs.get("bucket_name")
        service_account_file = gcp_configs.get("service_account_file")
        return GoogleCloudStorage(project_id=project_id, 
                bucket_name=bucket_name,
                service_account_file=service_account_file)
    if provider == "azure":
        # Set logging level
        connection_string = azure_configs.get("connection_string") 
        container_name = azure_configs.get("container_name")
        log_level = int(azure_configs.get("log_level"))
        logging.getLogger("azure.storage.common.storageclient")\
                .setLevel(log_level)
        return AzureStorage(connection_string=connection_string,
                container_name=container_name)
    raise ValueError("Uknown provider: "+provider)



import os

import psycopg2
from psycopg2.extras import Json, RealDictCursor
from celery.utils.log import get_task_logger
import spacy
from bs4 import BeautifulSoup
import requests

from .celery import app
from .storage.objects import storage
from .settings import db_configs
from .models.language_models import LanguageModel as lm


logger = get_task_logger(__name__)


@app.task(ignore_result=True)
def index_ckan_package(
        crawler_package_key,
        package_blob_name,
        endpoint):
    """Register the crawled CKAN package into the centralized packages table for
    all types of packages to make it searchable. The following processes are
    performed:
        1. Extract metadata such as title an description from the raw package
            JSON file.
        2. Create the fulltext search doc for keyword search.
        3. Extract named entities from the title and description for
            metadata enrichment.
        4. Submit separate jobs for indexing ckan package files and
            sketching.
    """
    # Get package metadata
    package = storage.get_object(package_blob_name)

    # Get connection.
    conn = psycopg2.connect(**db_configs)
    # Get CKAN resources
    cur = conn.cursor(cursor_factory=RealDictCursor)
    cur.execute("""SELECT key, filename, resource_blob, file_size, raw_metadata
            FROM findopendata.ckan_resources where package_key = %s""",
            (crawler_package_key,))
    resources = [row for row in cur]

    # Parse package metadata
    created = package.get("metadata_created", None)
    modified = package.get("metadata_modified", None)
    title = BeautifulSoup(package.get("title", ""), "html.parser").get_text()
    title_spacy = lm.process(title).to_json()
    name = package.get("name", None)
    description = BeautifulSoup(package.get("notes", ""), "html.parser").get_text()
    description_spacy = lm.process(description).to_json()
    tags = [tag["name"] for tag in package.get("tags", []) if "name" in tag]
    license_title = package.get("license_title", None)
    license_url = package.get("license_url", None)
    organization_display_name = package.get("organization", {})\
            .get("title", None)
    organization_name = package.get("organization", {}).get("name", None)
    if organization_name is not None:
        organization_name = "{}:{}".format(endpoint, organization_name)
    organization_image_url = package.get("organization", {})\
            .get("image_url", None)
    organization_description = package.get("organization", {})\
            .get("description", None)
    raw_metadata = package

    crawler_table = "ckan_packages"
    crawler_key = crawler_package_key
    original_host = endpoint
    num_files = len(resources)

    # Save package
    cur.execute(r"""INSERT INTO findopendata.packages
            (
                crawler_table,
                crawler_key,
                id,
                original_host,
                num_files,
                created,
                modified,
                title,
                title_spacy,
                name,
                description,
                description_spacy,
                tags,
                license_title,
                license_url,
                organization_display_name,
                organization_name,
                organization_image_url,
                organization_description,
                raw_metadata
            )
            VALUES (
                %s, %s, uuid_generate_v1mc(), %s, %s,
                %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s
            )
            ON CONFLICT (crawler_table, crawler_key) DO UPDATE SET
            original_host = EXCLUDED.original_host,
            num_files = EXCLUDED.num_files,
            updated = current_timestamp,
            created = EXCLUDED.created,
            modified = EXCLUDED.modified,
            title = EXCLUDED.title,
            title_spacy = EXCLUDED.title_spacy,
            name = EXCLUDED.name,
            description = EXCLUDED.description,
            description_spacy = EXCLUDED.description_spacy,
            tags = EXCLUDED.tags,
            license_title = EXCLUDED.license_title,
            license_url = EXCLUDED.license_url,
            organization_display_name = EXCLUDED.organization_display_name,
            organization_name = EXCLUDED.organization_name,
            organization_image_url = EXCLUDED.organization_image_url,
            organization_description = EXCLUDED.organization_description,
            raw_metadata = EXCLUDED.raw_metadata
            RETURNING key;""",
            (
                crawler_table,
                crawler_key,
                original_host,
                num_files,
                created,
                modified,
                title,
                Json(title_spacy),
                name,
                description,
                Json(description_spacy),
                tags,
                license_title,
                license_url,
                organization_display_name,
                organization_name,
                organization_image_url,
                organization_description,
                Json(raw_metadata),
            ))
    package_key = cur.fetchone()["key"]
    # Commit all changes.
    conn.commit()
    conn.close()
    logger.info("Indexed CKAN package {} and {} package files".format(
        crawler_package_key, len(resources)))

    # Submit jobs for indexing package files.
    for resource in resources:
        index_ckan_package_file(crawler_resource_key=resource["key"],
                package_key=package_key,
                blob_name=resource["resource_blob"],
                filename=resource["filename"],
                file_size=resource["file_size"],
                raw_metadata=resource["raw_metadata"])


@app.task(ignore_result=True)
def index_ckan_package_file(
        crawler_resource_key,
        package_key,
        blob_name,
        filename,
        file_size,
        raw_metadata):
    """Register the CKAN resource into the package_files table by doing the
    following:
        1. Extract metadata such as name and description from the source JSON.

    Args:
        crawler_resource_key: the primary key of the ckan_resources table.
        package_key: the primary key of the packages table.
        blob_name: the relative path to the blob of this package file.
        filename: the UNIX-friendly filename.
        file_size: the file size in bytes.
        raw_metadata: the 'resource' extracted from the CKAN's package JSON
            corresponding to this package file.
    """
    # Extract metadata from raw_metadata
    crawler_table = "ckan_resources"
    created = raw_metadata.get("created", None)
    modified = raw_metadata.get("last_modified", None)
    name = raw_metadata.get("name", None)
    description = BeautifulSoup(raw_metadata.get("description", ""),
            "html.parser").get_text()
    description_spacy = lm.process(description).to_json()
    original_url = raw_metadata.get("url", None)
    file_format = raw_metadata.get("format", None)

    # Get connection.
    conn = psycopg2.connect(**db_configs)
    # Save CKAN pacakge file
    cur = conn.cursor(cursor_factory=RealDictCursor)
    cur.execute(r"""INSERT INTO findopendata.package_files
            (
                package_key,
                crawler_table,
                crawler_key,
                id,
                created,
                modified,
                filename,
                name,
                description,
                description_spacy,
                original_url,
                format,
                file_size,
                blob_name,
                raw_metadata
            ) VALUES
            (%s, %s, %s, uuid_generate_v1mc(), %s, %s,
            %s, %s, %s, %s, %s, %s, %s, %s,
            %s) ON CONFLICT (crawler_table, crawler_key) DO UPDATE SET
            updated = current_timestamp,
            created = EXCLUDED.created,
            modified = EXCLUDED.modified,
            filename = EXCLUDED.filename,
            name = EXCLUDED.name,
            description = EXCLUDED.description,
            description_spacy = EXCLUDED.description_spacy,
            original_url = EXCLUDED.original_url,
            format = EXCLUDED.format,
            file_size = EXCLUDED.file_size,
            blob_name = EXCLUDED.blob_name,
            raw_metadata = EXCLUDED.raw_metadata
            RETURNING key;""",
            (
                package_key,
                crawler_table,
                crawler_resource_key,
                created,
                modified,
                filename,
                name,
                description,
                Json(description_spacy),
                original_url,
                file_format,
                file_size,
                blob_name,
                Json(raw_metadata)
            ))
    # Commit all changes.
    conn.commit()
    conn.close()
    logger.info("Indexed CKAN resource {} into package_files".format(
        crawler_resource_key))


@app.task(ignore_result=True)
def index_socrata_resource(
        crawler_key,
        domain,
        metadata_blob_name,
        resource_blob_name,
        dataset_size):
    """Register the crawled Scorata resource into the centralized packages
    table for all types of packages to make it searchable.
    The following processes are performed:
        1. Extract metadata such as title and description from the raw metadata
            JSON file.
        2. Create the fulltext search doc for keyword search.
        3. Extract named entities from the title and description for
            metadata enrichment.
    """
    # Get raw metadata
    metadata = storage.get_object(metadata_blob_name)

    # Package fields extracted from raw metadata.
    created = metadata["resource"].get("createdAt")
    modified = metadata["resource"].get("updatedAt")
    title = metadata["resource"].get("name")
    title_spacy = None
    name = None
    description = metadata["resource"].get("description")
    description_spacy = None
    tags = metadata["classification"].get("domain_tags")
    license_title = metadata["metadata"].get("license")
    license_url = None
    organization_display_name = metadata["resource"].get("attribution")
    organization_name = None
    organization_image_url = None
    organization_description = None

    # Package fields assigned
    crawler_table = "socrata_resources"
    original_host = domain
    num_files = 1

    # Get connection.
    conn = psycopg2.connect(**db_configs)
    # Get CKAN resources
    cur = conn.cursor(cursor_factory=RealDictCursor)

    # Save package
    cur.execute(r"""INSERT INTO findopendata.packages
            (
                crawler_table,
                crawler_key,
                id,
                original_host,
                num_files,
                created,
                modified,
                title,
                title_spacy,
                name,
                description,
                description_spacy,
                tags,
                license_title,
                license_url,
                organization_display_name,
                organization_name,
                organization_image_url,
                organization_description,
                raw_metadata
            )
            VALUES (
                %s, %s, uuid_generate_v1mc(), %s, %s,
                %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s
            )
            ON CONFLICT (crawler_table, crawler_key) DO UPDATE SET
            original_host = EXCLUDED.original_host,
            num_files = EXCLUDED.num_files,
            updated = current_timestamp,
            created = EXCLUDED.created,
            modified = EXCLUDED.modified,
            title = EXCLUDED.title,
            title_spacy = EXCLUDED.title_spacy,
            name = EXCLUDED.name,
            description = EXCLUDED.description,
            description_spacy = EXCLUDED.description_spacy,
            tags = EXCLUDED.tags,
            license_title = EXCLUDED.license_title,
            license_url = EXCLUDED.license_url,
            organization_display_name = EXCLUDED.organization_display_name,
            organization_name = EXCLUDED.organization_name,
            organization_image_url = EXCLUDED.organization_image_url,
            organization_description = EXCLUDED.organization_description,
            raw_metadata = EXCLUDED.raw_metadata
            RETURNING key;""",
            (
                crawler_table,
                crawler_key,
                original_host,
                num_files,
                created,
                modified,
                title,
                Json(title_spacy),
                name,
                description,
                Json(description_spacy),
                tags,
                license_title,
                license_url,
                organization_display_name,
                organization_name,
                organization_image_url,
                organization_description,
                Json(metadata),
            ))
    package_key = cur.fetchone()["key"]

    # Package file fields extracted
    created = metadata["resource"].get("createdAt")
    modified = metadata["resource"].get("data_updated_at")
    original_url = metadata["link"]
    file_format = 'AVRO'
    file_size = dataset_size
    blob_name = resource_blob_name
    raw_metadata = metadata["resource"]

    # Save pacakge file:
    cur.execute(r"""INSERT INTO findopendata.package_files
            (
                package_key,
                crawler_table,
                crawler_key,
                id,
                created,
                modified,
                original_url,
                format,
                file_size,
                blob_name,
                raw_metadata
            ) VALUES
            (%s, %s, %s, uuid_generate_v1mc(), %s,
            %s, %s, %s, %s, %s, %s)
            ON CONFLICT (crawler_table, crawler_key) DO UPDATE SET
            updated = current_timestamp,
            created = EXCLUDED.created,
            modified = EXCLUDED.modified,
            original_url = EXCLUDED.original_url,
            format = EXCLUDED.format,
            file_size = EXCLUDED.file_size,
            blob_name = EXCLUDED.blob_name,
            raw_metadata = EXCLUDED.raw_metadata
            RETURNING key;""",
            (
                package_key,
                crawler_table,
                crawler_key,
                created,
                modified,
                original_url,
                file_format,
                file_size,
                blob_name,
                Json(raw_metadata)
            ))

    # Commit all changes.
    conn.commit()
    conn.close()
    logger.info("Indexed Socrata resource {}".format(crawler_key))



import os

import psycopg2
from psycopg2.extras import Json, RealDictCursor, register_uuid
import dateutil.parser
from celery.utils.log import get_task_logger

from .celery import app
from .settings import db_configs
from .storage.objects import storage
from .parsers.csv import csv2json
from .parsers.avro import avro2json
from .parsers.jsonl import jsonl2json
from .column_sketch import ColumnSketch
from .table_sketch import TableSketch


logger = get_task_logger(__name__)


def _json_records_sketcher(records, record_sample_size=20, max_records=None,
        **kwargs):
    table_sketch = TableSketch(record_sample_size=record_sample_size, **kwargs)
    if max_records is not None:
        records = (record for record, _ in zip(records, range(max_records)))
    for record in records:
        table_sketch.update(record)
    return table_sketch


def _csv_sketcher(fileobj_binary, record_sample_size=20, max_records=None,
        **kwargs):
    records = csv2json(fileobj_binary)
    return _json_records_sketcher(records, record_sample_size, max_records,
        **kwargs)


def _jsonl_sketcher(fileobj_binary, record_sample_size=20, max_records=None,
        **kwargs):
    records = jsonl2json(fileobj_binary)
    return _json_records_sketcher(records, record_sample_size, max_records,
        **kwargs)


def _avro_sketcher(fileobj_binary, record_sample_size=20, max_records=None,
        **kwargs):
    records = avro2json(fileobj_binary)
    return _json_records_sketcher(records, record_sample_size, max_records,
        **kwargs)


_sketchers = {
        "csv": _csv_sketcher,
        "jsonl": _jsonl_sketcher,
        "avro": _avro_sketcher,
        }


@app.task(ignore_result=True)
def sketch_package_file(package_file_key,
        blob_name,
        dataset_format,
        max_records,
        table_sample_size,
        minhash_size,
        minhash_seed,
        hyperloglog_p,
        column_sample_size,
        enable_word_vector_data):
    """Generate column sketches and table sample of the table in the
    package file.

    Args:
        package_file_key: the primary key of package_files table.
        blob_name: the relative path to the blob of the package file.
        dataset_format: one of csv, jsonl, and avro.
        max_records: the maximum number of records to sketch.
        table_sample_size: the number of records include in the table sample.
        minhash_size: the number of permutation (hash functions) to use for
            MinHash sketches.
        minhash_seed: the random seed for generating MinHash sketches'
            permutations.
        hyperloglog_p: the precision parameter used by HyperLogLog.
        column_sample_size: the number of non-random sampled values.
        enable_word_vector_data: whether to create word vectors for the
            data values -- this can be 10x more expensive.
    """
    # Get sketcher
    if dataset_format not in _sketchers:
        raise ValueError("{} is not supported".format(dataset_format))
    sketcher = _sketchers[dataset_format]

    # Sketch the file.
    try:
        with storage.get_file(blob_name) as input_file:
            table_sketch = sketcher(input_file,
                    record_sample_size=table_sample_size,
                    max_records=max_records,
                    minhash_size=minhash_size,
                    minhash_seed=minhash_seed,
                    hyperloglog_p=hyperloglog_p,
                    sample_size=column_sample_size,
                    enable_word_vector_data=enable_word_vector_data
                    )
    except Exception as e:
        logger.error("Sketching {} ({}) failed due to {}".format(
            blob_name, package_file_key, e))
        raise e

    try:
        # Save sketches to the database
        # Initialize Postgres connection.
        conn = psycopg2.connect(**db_configs)
        cur = conn.cursor(cursor_factory=RealDictCursor)
        register_uuid(conn_or_curs=cur)
        # Save column sketches
        column_sketch_ids = []
        for sketch in table_sketch.column_sketches:
            cur.execute(r"""INSERT INTO findopendata.column_sketches
                    (
                        package_file_key,
                        id,
                        column_name,
                        sample,
                        count,
                        empty_count,
                        out_of_vocabulary_count,
                        numeric_count,
                        distinct_count,
                        word_vector_column_name,
                        word_vector_data,
                        minhash,
                        seed,
                        hyperloglog
                    )
                    VALUES (%s, uuid_generate_v1mc(),
                        %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s)
                    ON CONFLICT (package_file_key, column_name)
                    DO UPDATE
                    SET updated = current_timestamp,
                    sample = EXCLUDED.sample,
                    count = EXCLUDED.count,
                    empty_count = EXCLUDED.empty_count,
                    out_of_vocabulary_count = EXCLUDED.out_of_vocabulary_count,
                    numeric_count = EXCLUDED.numeric_count,
                    distinct_count = EXCLUDED.distinct_count,
                    word_vector_column_name = EXCLUDED.word_vector_column_name,
                    word_vector_data = EXCLUDED.word_vector_data,
                    minhash = EXCLUDED.minhash,
                    seed = EXCLUDED.seed,
                    hyperloglog = EXCLUDED.hyperloglog
                    RETURNING id::uuid
                    """, (
                        package_file_key,
                        sketch.column_name,
                        sketch.sample,
                        sketch.count,
                        sketch.empty_count,
                        sketch.out_of_vocabulary_count,
                        sketch.numeric_count,
                        sketch.distinct_count,
                        sketch.word_vector_column_name,
                        sketch.word_vector_data,
                        sketch.minhash,
                        sketch.seed,
                        sketch.hyperloglog,
                        ))
            column_sketch_ids.append(cur.fetchone()["id"])
        # Save table samples, column names and column sketch IDs.
        cur.execute(r"""UPDATE findopendata.package_files
                        SET column_names = %s,
                        column_sketch_ids = %s,
                        sample = %s
                        WHERE key = %s
                        """, (
                            table_sketch.column_names,
                            column_sketch_ids,
                            Json(table_sketch.record_sample),
                            package_file_key,
                            ))
        # Commit
        conn.commit()
        cur.close()
        conn.close()
    except Exception as e:
        logger.error("Error saving sketches of {} ({}) due to {}".format(
            blob_name, package_file_key, e))
        raise e

    # Finish
    logger.info("Sketching {} ({}) successful".format(blob_name,
        package_file_key))




import os

import yaml


def from_yaml(filename):
    with open(filename, "r") as f:
        return yaml.load(f, Loader=yaml.SafeLoader)


# Configurations.
configs = from_yaml(os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    os.path.pardir,
    "configs.yaml"
))

# Storage configurations.
storage_configs = configs.get("storage")

# Local storage configurations.
local_configs = configs.get("local")

# GCP configurations.
gcp_configs = configs.get("gcp")

# Azure configurations.
azure_configs = configs.get("azure")

# Crawler configurations.
crawler_configs = configs.get("crawler")

# Postgres configurations.
db_configs = configs.get("postgres")

# Celery configurations.
celery_configs = configs.get("celery")

# Index configurations.
index_configs = configs.get("index")



import abc
import contextlib


class Blob(object):

    def __init__(self, name: str, size: int):
        self._name = name
        self._size = size
    
    @property
    def name(self) -> str:
        return self._name
    
    @property
    def size(self) -> int:
        return self._size


class BlobStorage(object):

    @abc.abstractmethod
    def get_object(self, blob_name):
        """Get a JSON blob as a Python dictionary object.

        Args:
            blob_name: the name of the blob.

        Returns: a Python dictionary.
        """
        pass

    @abc.abstractmethod
    @contextlib.contextmanager
    def get_file(self, blob_name):
        """Get a blob as an opened binary file for readonly.

        Args:
            blob_name: the name of the blob.
        
        Returns: a binary file object.

        Example:

            with storage.get_file(blob_name) as f:
                # Do stuff
        """
        pass

    @abc.abstractmethod
    def put_file(self, fileobj, blob_name):
        """Save a flat file to the storage.

        Args:
            fileobj: the file object to be uploaded to the bucket.
            blob_name: the name of the destination blob.

        Returns:
            blob: the blob object that was created.
        """
        pass

    @abc.abstractmethod
    def put_object(self, obj, blob_name):
        """Save a single JSON-serializable Python dictionary to storage.

        Args:
            obj: the Python dictionary to be JSON-serialized and saved to 
                storage.
            blob_name: the name of the destination blob.

        Returns:
            blob: the blob object that was created.
        """
        pass

    @abc.abstractmethod
    def put_avro(self, schema, records, blob_name, codec="snappy"):
        """Save Avro records to storage.

        Args:
            schema: the Avro schema to use.
            records: an iterator of records of type `dict`.
            blob_name: the name of the destination blob.
            codec: the compression codec to use (e.g., `snappy`).

        Returns:
            blob: the blob object that was created.
        """
        pass

    @abc.abstractmethod
    def put_json(self, records, blob_name, gzip_compress=True):
        """Save JSON records to storage as a newline-delimited JSON file.

        Args:
            records: an iterator of records of type `dict`.
            blob_name: the name of the destination blob.
            gzip_compress: whether to use Gzip compression for the blob.

        Returns:
            blob: the blob object that was created.
        """
        pass



from . import _SpacyModel

# When _LazySpacyModel is used, importing the model won't immediately load
# the model, however every Celery worker process will load a redundant copy of
# the model into its own memory -- high memory usage.
# WordVectorModel = _LazySpacyModel("en_vectors_web_lg")

# When _SpacyModel is used, the parent Celery process will load the model
# into its memory, and Celery workers (forked child processes) may not need
# to have physical redundant copies due to "copy-on-write".
WordVectorModel = _SpacyModel("en_core_web_sm")



import unittest
import random
import cProfile

from findopendata.column_sketch import ColumnSketch
from findopendata.models.word_vector_models import WordVectorModel as lm


WORDS = [
    "english",
    "french",
    "spanish",
    "russian",
    "italian",
    "arabic",
    "sanskrit",
    "chinese",
    "japanese",
    "korean",
]

TEST_COLUMN_1_NAME = "Countries"
TEST_COLUMN_1 = WORDS * 10000


class TestColumnSketch(unittest.TestCase):
    def setUp(self):
        self.pr = cProfile.Profile()
        self.pr.enable()

    def tearDown(self):
        self.pr.disable()
        self.pr.dump_stats("{}.prof".format(self._testMethodName))

    @unittest.skip("")
    def test_disable_word_vec_data(self):
        sketch = ColumnSketch(
            TEST_COLUMN_1_NAME,
            minhash_size=256,
            minhash_seed=43,
            hyperloglog_p=8,
            sample_size=100,
            enable_word_vector_data=False,
            model=lm,
        )
        for word in TEST_COLUMN_1:
            sketch.update(word)
        self.assertTrue(sketch.count == len(TEST_COLUMN_1))
        self.assertTrue(len(sketch.sample) == len(WORDS))
        self.assertTrue(sketch.word_vector_column_name is not None)
        self.assertTrue(sketch.word_vector_data is None)
        self.assertTrue(sketch.non_empty_count == len(TEST_COLUMN_1))
        self.assertTrue(sketch.hyperloglog is not None)
        self.assertTrue(len(sketch.minhash) == 256)

    @unittest.skip("")
    def test_enable_word_vec_data(self):
        sketch = ColumnSketch(
            TEST_COLUMN_1_NAME,
            minhash_size=256,
            minhash_seed=43,
            hyperloglog_p=8,
            sample_size=100,
            enable_word_vector_data=True,
            model=lm,
        )
        for word in TEST_COLUMN_1:
            sketch.update(word)
        self.assertTrue(sketch.count == len(TEST_COLUMN_1))
        self.assertTrue(len(sketch.sample) == len(WORDS))
        self.assertTrue(sketch.word_vector_column_name is not None)
        self.assertTrue(sketch.word_vector_data is not None)
        self.assertTrue(sketch.non_empty_count == len(TEST_COLUMN_1))
        self.assertTrue(sketch.hyperloglog is not None)
        self.assertTrue(len(sketch.minhash) == 256)


if __name__ == "__main__":
    unittest.main()



import os
import contextlib
import gzip

import simplejson as json
import fastavro
from google.oauth2 import service_account
from google.cloud import storage
from gcsfs.core import GCSFileSystem

from .base import BlobStorage, Blob


class GoogleCloudStorage(BlobStorage):
    """Google Cloud Storage storage provider.

    Args:
        project_id: the ID of the Google Cloud project.
        bucket_name: the name of the Cloud Storage bucket to use for all blobs.
        service_account_file: the filename of the GCP service account JSON key 
            file.
    """
    def __init__(self, project_id: str, bucket_name: str, 
            service_account_file: str):
        self._bucket_name = bucket_name
        self._client = storage.Client(project=project_id, 
                credentials=service_account.Credentials.\
                from_service_account_file(service_account_file))
        self._fs = GCSFileSystem(token=service_account_file,
                check_connection=True)
    
    def get_object(self, blob_name):
        blob = self._client.bucket(self._bucket_name).get_blob(blob_name)
        if blob is None:
            raise ValueError("Cannot find blob: "+blob_name)
        return json.loads(blob.download_as_string().decode("utf-8"))

    @contextlib.contextmanager
    def get_file(self, blob_name):
        path = os.path.join(self._bucket_name, blob_name)
        try:
            fileobj = self._fs.open(path, 'rb')
            yield fileobj
        finally:
            fileobj.close()

    def put_object(self, obj, blob_name):
        blob = self._client.bucket(self._bucket_name).blob(blob_name)
        data = json.dumps(obj).encode("utf-8")
        blob.upload_from_string(data, content_type="application/json")
        blob.reload()
        return Blob(blob_name, blob.size)

    def put_file(self, fileobj, blob_name):
        blob = self._client.bucket(self._bucket_name).blob(blob_name)
        blob.upload_from_file(fileobj)
        blob.reload()
        return Blob(blob_name, blob.size)
    
    def put_avro(self, schema, records, blob_name, codec='snappy'):
        path = os.path.join(self._bucket_name, blob_name)
        tmp_path = os.path.join(os.path.dirname(path),
                "~{}".format(os.path.basename(path)))
        with self._fs.open(tmp_path, "wb") as of:
            fastavro.writer(of, schema, records, codec)
        self._fs.mv(tmp_path, path)
        self._fs.setxattrs(path, content_type="avro/binary")
        blob = self._client.bucket(self._bucket_name).get_blob(blob_name)
        if blob is None:
            raise RuntimeError("Cannot find new avro blob: "+blob_name)
        return Blob(blob_name, blob.size)
    
    def put_json(self, records, blob_name, gzip_compress=True):
        path = os.path.join(self._bucket_name, blob_name)
        tmp_path = os.path.join(os.path.dirname(path),
                "~{}".format(os.path.basename(path)))
        newline = "\n"
        with self._fs.open(tmp_path, "wb") as of:
            if gzip_compress:
                with gzip.open(of, "wt") as of:
                    for record in records:
                        of.write(json.dumps(record))
                        of.write(newline)
            else:
                for record in records:
                    of.write(json.dumps(record))
                    of.write(newline)
        self._fs.mv(tmp_path, path)
        self._fs.setxattrs(path, content_type="application/json")
        blob = self._client.bucket(self._bucket_name).get_blob(blob_name)
        if blob is None:
            raise RuntimeError("Cannot find new JSON blob: "+blob_name)
        return Blob(blob_name, blob.size)



import re
import io
import random

import requests
import dateutil.parser
import psycopg2
from celery.utils.log import get_task_logger

from .celery import app
from .storage.objects import storage
from .parsers.avro import JSON2AvroRecords
from .settings import crawler_configs, db_configs, gcp_configs
from .socrata import socrata_records


# Logger for tasks.
logger = get_task_logger(__name__)


@app.task(ignore_result=True)
def add_socrata_resource(metadata, app_token, blob_prefix, force_update):
    """Retrieves and adds a Socrata resource to the registry.

    Args:
        metadata: the 'resource' section of the JSON data returned from
            the Socrata Discovery API.
        app_token: the application token for using the discovery API.
        blob_prefix: the prefix for all the blobs uploaded from this
            function, relative to the root.
        force_update: whether to force update regardless of the updated time.
    """

    domain = metadata["metadata"]["domain"]
    try:
        uid, original_url = _extract_socrata_uid2(metadata)
    except Exception as e:
        logger.warning("(domain={} Failed to extract UID and resource URL from "
                "{}: {}".format(domain, metadata["permalink"], e))
        return

    if not force_update:
        # Initialize Postgres connection and cursor.
        conn = psycopg2.connect(**db_configs)
        cur = conn.cursor()

        # Check if the resource already exists and is updated.
        cur.execute("SELECT updated::timestamptz "
                "FROM findopendata.socrata_resources "
                "WHERE domain = %s AND id = %s;", (domain, uid))
        row = cur.fetchone()
        if row is not None:
            last_registered = row[0]
            last_updated = dateutil.parser.parse(
                    metadata["resource"]["updatedAt"])
            if last_updated <= last_registered:
                logger.info("(domain={} id={}) Skipping (updated {}, "
                        "registered {})".format(domain, uid, last_updated,
                            last_registered))
                return

        # Close the database connection to prevent download hogging the pool.
        cur.close()
        conn.close()

    # Upload the metadata.
    metadata_blob = storage.put_object(metadata, 
            "/".join([blob_prefix, domain, uid, "metadata.json"]))
    logger.info("(domain={} id={}) Saved metadata.".format(domain, uid))

    # Get the field names from metadata.
    field_names = metadata["resource"]["columns_field_name"]

    # Download and upload the resource.
    logger.info("(domain={} id={}) Saving resource from {}.".format(
        domain, uid, original_url))
    resource_blob_name = "/".join([blob_prefix, domain, uid, "resource.avro"])
    try:
        records = JSON2AvroRecords(socrata_records(original_url, app_token),
                field_names=field_names)
        resource_blob = storage.put_avro(records.schema, records.get(),
                resource_blob_name, codec="snappy")
    except Exception as e:
        logger.warning("(domain={} id={}) Failed to save resource from {}: {}".\
                format(domain, id, original_url, e))
        return
    logger.info("(domain={} id={}) Finished saving resource from {} to {}".\
            format(domain, uid, original_url, resource_blob_name))

    # Initialize Postgres connection and cursor for registering resource.
    conn = psycopg2.connect(**db_configs)
    cur = conn.cursor()

    # Register this resource.
    cur.execute("INSERT INTO findopendata.socrata_resources "
            "(domain, id, metadata_blob, resource_blob, original_url, "
            "dataset_size) VALUES (%s, %s, %s, %s, %s, %s) "
            "ON CONFLICT (domain, id) DO UPDATE "
            "SET updated = current_timestamp, "
            "metadata_blob = EXCLUDED.metadata_blob, "
            "resource_blob = EXCLUDED.resource_blob, "
            "original_url = EXCLUDED.original_url, "
            "dataset_size = EXCLUDED.dataset_size;",
            (domain, uid, metadata_blob.name, resource_blob.name,
                original_url, resource_blob.size))
    conn.commit()

    # Close database connection for checking resource.
    cur.close()
    conn.close()

    logger.info("(domain={} id={}) Successful.".format(domain, uid))


def _extract_socrata_uid2(metadata):
    # Obtain the resource URL and UID (SODA V2.1) from the web page
    # TODO: check with Scorata see if a fix has been put forward
    # to resolve the 2.0 UID into 2.1 UID.
    # https://github.com/socrata/cetera/issues/260
    domain = metadata["metadata"]["domain"]
    weburl = metadata["permalink"]
    resource_url_reg = re.compile(r"(https?:\/\/" + domain + \
            r"\/resource\/[a-z0-9A-Z]{4}-[a-z0-9A-Z]{4}\.json)")
    uid_reg = re.compile(r"https?:\/\/" + domain + \
            r"\/resource\/([a-z0-9A-Z]{4}-[a-z0-9A-Z]{4})\.json")
    resp = requests.get(weburl)
    resp.raise_for_status()
    items = resource_url_reg.findall(resp.text)
    if len(items) != 1:
        raise RuntimeError("%d resource URL is found from %s: "
                "%s, expecting 1" % (len(items), weburl, items))
    resource_url = items[0]
    uid = uid_reg.findall(resource_url)[0]
    return uid, resource_url


@app.task(ignore_result=True)
def add_socrata_resources_from_api(discovery_api_url, blob_prefix,
        force_update):
    """Scrolls through the Socrata Discovery API and starts tasks to retrieves
    resources.

    Args:
        discovery_api_url: the full URL of the Discovery API endpoint.
        blob_prefix: the prefix for all the blobs uploaded from this
            function, relative to the root.
    """
    logger.info("(discovery_api_url=%s)" % (discovery_api_url))
    app_tokens = _get_socrata_app_tokens()
    sources = _process_socrata_raw_metadata(discovery_api_url, 50, 
            random.choice(app_tokens))
    for source in sources:
        add_socrata_resource.delay(source, app_token=random.choice(app_tokens),
                blob_prefix=blob_prefix, force_update=force_update)


def _process_socrata_raw_metadata(discovery_api_url, page_size, token):
    scroll_id = ""
    sess = requests.session()
    while True:
        resp = sess.get(discovery_api_url,
                params={"scroll_id" : scroll_id,
                        "limit" : page_size,
                        "provenance" : "official",
                        "only" : "datasets"},
                headers={"X-App-Token" : token})
        resp.raise_for_status()
        results = resp.json()["results"]
        if len(results) == 0:
            break
        scroll_id = results[-1]["resource"]["id"]
        for metadata in results:
            yield metadata


def _get_socrata_app_tokens():
    conn = psycopg2.connect(**db_configs)
    cur = conn.cursor()
    cur.execute("SELECT token FROM findopendata.socrata_app_tokens "
            "ORDER BY random()")
    rows = cur.fetchall()
    cur.close()
    conn.close()
    if not rows:
        raise RuntimeError("Cannot found a Socrata app token!")
    return [row[0] for row in rows]


@app.task(ignore_result=True)
def add_socrata_discovery_apis(force_update):
    """Add Socrata Discovery API endpoints to the crawler."""
    conn = psycopg2.connect(**db_configs)
    cur = conn.cursor()
    cur.execute("SELECT url "
            "FROM findopendata.socrata_discovery_apis WHERE enabled = true")
    api_urls = [row[0] for row in cur]
    cur.close()
    conn.close()
    for api_url in api_urls:
        add_socrata_resources_from_api.delay(api_url,
                crawler_configs.get("socrata_blob_prefix"), force_update)
        logger.info("Adding Socrata Discovery API {} to the crawler".format(
            api_url))



import requests


def socrata_records(resource_url, app_token, limit=25000):
    """Reading records from given resource URL.

    Args:
        resource_url: the Socrata dataset API.
        app_token: the Socrata dataset access credential.
        limit (defautl 25000): the pagination limit used for reading the
            Socrata dataset API.
    """
    def _call_api(resource_url, app_token, limit, offset):
        if app_token is not None:
            resp = requests.get(resource_url,
                    params={
                        r"$order" : r":id",
                        r"$offset" : offset,
                        r"$limit" : limit},
                    headers={
                        "X-App-Token" : app_token})
            if resp.status_code == 403:
                # Re-try without app token
                return _call_api(resource_url, None, limit, offset)
        else:
            resp = requests.get(resource_url,
                    params={
                        r"$order" : r":id",
                        r"$offset" : offset,
                        r"$limit" : limit})
        if not resp.ok:
            resp.raise_for_status()
        records = resp.json()
        return records

    # Main loop
    offset = 0
    while True:
        records = _call_api(resource_url, app_token, limit, offset)
        offset += limit
        if len(records) == 0:
            return
        for record in records:
            yield record



import cchardet as chardet


def guess_encoding_from_buffer(buf, chardet_threshold=0.5):
    result = chardet.detect(buf)
    confidence = result.get("confidence")
    if not confidence or confidence < chardet_threshold:
        raise ValueError("Failed to detect encoding")
    encoding = result["encoding"]
    return encoding


def guess_encoding_from_stream(stream, chunk_size=4096, chardet_threshold=0.5):
    detector = chardet.UniversalDetector()
    chunk = stream.read(chunk_size)
    while not detector.done and chunk:
        detector.feed(chunk)
        chunk = stream.read(chunk_size)
    detector.close()
    result = detector.result
    confidence = result.get("confidence")
    if not confidence or confidence < chardet_threshold:
        raise ValueError("Failed to detect encoding")
    encoding = result["encoding"]
    return encoding



import io
import unittest
from collections import OrderedDict

import fastavro

from findopendata.parsers.avro import JSON2AvroRecords, avro2json


records = [
        {"username": "javasucks", "email": "go@example.com"},
        {"username": "moonshoot", "email": "moon@example.com"},
        {"username": "twilight", "email": "tw@example.com"},
        {"username": "birdeye", "email": "be@example.com", "amount": 0},
        {"username": "birdeye", "amount": 1000},
        ]

record_nested = [
    {
        "username": "javasucks", 
        "email": "go@example.com",
        "location": {
            "latitude": 12.22,
            "longitude": -84.23,
        },
    },
    {
        "username": "moonshoot", 
        "email": "moon@example.com",
        "location": {
            "latitude": 11.22,
            "longitude": -54.23,
        },
    },
    {
        "username": "twilight", 
        "email": "tw@example.com",
        "location": {
            "latitude": 19.22,
            "longitude": -80.23,
        },
    },
    {
        "username": "birdeye", 
        "email": "be@example.com", 
        "amount": 0,
        "location": {
            "latitude": 120.22,
            "longitude": 23.23,
        },
     },
    {
        "username": "birdeye", 
        "amount": 1000,
        "location": {
            "latitude": 22.22,
            "longitude": 30.23,
        },
    },
]

field_names = ["email", "username", "amount"]

field_names_nested = ["email", "username", "amount", "location"]


class TestJSON2AvroRecords(unittest.TestCase):

    def test_basic(self):
        test_records = JSON2AvroRecords((r for r in records))
        self.assertEqual(len(list(test_records.get())), 5)
        schema = test_records.schema
        self.assertEqual(len(schema["fields"]), 3)

    def test_field_order(self):
        test_records = JSON2AvroRecords((r for r in records),
                field_names=field_names)
        schema = test_records.schema
        self.assertEqual([f["name"] for f in schema["fields"]], field_names)

    def test_nested(self):
        test_records = JSON2AvroRecords((r for r in record_nested),
                field_names=field_names)
        schema = test_records.schema
        self.assertEqual([f["name"] for f in schema["fields"]], 
                field_names_nested)


class TestAvro2JSONRecords(unittest.TestCase):

    def test_avro2json(self):
        test_records = JSON2AvroRecords((r for r in records),
                field_names=field_names)
        buf = io.BytesIO(b'')
        fastavro.writer(buf, test_records.schema, test_records.get())
        buf.seek(0)
        for record in avro2json(buf):
            self.assertTrue(isinstance(record, OrderedDict))
            self.assertEqual(list(record.keys()), field_names)


if __name__ == "__main__":
    unittest.main()




#!/usr/bin/env python
import sys
import argparse
import collections

import psycopg2

from findopendata.settings import db_configs
from findopendata.metadata import index_ckan_package, index_socrata_resource


_sql_ckan_force_update = r"""
SELECT p.key, p.package_blob, p.endpoint
FROM findopendata.original_hosts as h, findopendata.ckan_packages as p
WHERE p.endpoint = h.original_host and h.enabled
"""

_sql_ckan = r"""
WITH updated_times AS (
    SELECT crawler_key as key, updated
    FROM findopendata.packages
    WHERE crawler_table = 'ckan_packages'
),
enabled_packages AS (
    SELECT p.key, p.package_blob, p.endpoint, p.updated
    FROM findopendata.original_hosts as h, findopendata.ckan_packages as p
    WHERE p.endpoint = h.original_host and h.enabled
),
packages AS (
    SELECT a.key, a.package_blob, a.endpoint, a.updated as crawler_updated,
        u.updated as package_updated
    FROM enabled_packages AS a
    LEFT JOIN updated_times AS u
    ON a.key = u.key
)
SELECT key, package_blob, endpoint
FROM packages
WHERE package_updated IS NULL OR crawler_updated > package_updated
"""

_sql_socrata_force_update = r"""
SELECT r.key, r.domain, r.metadata_blob, r.resource_blob, r.dataset_size
FROM findopendata.socrata_resources as r, findopendata.original_hosts as h
WHERE r.domain = h.original_host AND h.enabled
"""

_sql_socrata = r"""
WITH updated_times AS (
    SELECT crawler_key as key, updated
    FROM findopendata.packages
    WHERE crawler_table = 'socrata_resources'
),
enabled_resources AS (
    SELECT r.key, r.domain, r.metadata_blob, r.resource_blob, r.dataset_size,
        r.updated
    FROM findopendata.socrata_resources as r, findopendata.original_hosts as h
    WHERE r.domain = h.original_host AND h.enabled
),
resources AS (
    SELECT a.key, a.domain, a.metadata_blob, a.resource_blob,
        a.dataset_size, a.updated as crawler_updated,
        u.updated as package_updated
    FROM enabled_resources as a
    LEFT JOIN updated_times as u
    ON a.key = u.key
)
SELECT key, domain, metadata_blob, resource_blob, dataset_size
FROM resources
WHERE package_updated IS NULL OR crawler_updated > package_updated
"""

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Extract and generate package metadata and "
        "make the packages searchable.")
    parser.add_argument("-u", "--force-update", action="store_true")
    args = parser.parse_args(sys.argv[1:])

    # CKAN
    print("Creating CKAN tasks (force_update = {})...".format(args.force_update))
    ckan_packages = collections.deque([])
    conn = psycopg2.connect(**db_configs)
    cur = conn.cursor()
    cur.execute((_sql_ckan_force_update if args.force_update else _sql_ckan))
    for row in cur:
        ckan_packages.append(row)
    cur.close()
    conn.close()
    print("Sending {} CKAN tasks to workers...".format(len(ckan_packages)))
    for key, package_blob, endpoint in ckan_packages:
        index_ckan_package.delay(
            crawler_package_key=key,
            package_blob_name=package_blob,
            endpoint=endpoint)
    print("Done sending CKAN tasks.")
    ckan_packages.clear()

    # Socrata
    print("Creating Socrata tasks (force_update = {})...".format(args.force_update))
    socrata_resources = collections.deque([])
    conn = psycopg2.connect(**db_configs)
    cur = conn.cursor()
    cur.execute((_sql_socrata_force_update if args.force_update else _sql_socrata))
    for row in cur:
        socrata_resources.append(row)
    cur.close()
    conn.close()
    print("Sending {} Socrata tasks to workers...".format(len(socrata_resources)))
    for key, domain, metadata_blob, resource_blob, dataset_size in \
            socrata_resources:
        index_socrata_resource.delay(
            crawler_key=key,
            domain=domain,
            metadata_blob_name=metadata_blob,
            resource_blob_name=resource_blob,
            dataset_size=dataset_size)
    print("Done sending Socrata tasks.")
    socrata_resources.clear()



"""This module has the initialized storage object.
"""

from ..settings import storage_configs
from .factory import BlobStorageFactory

_provider = storage_configs.get("provider")
storage = BlobStorageFactory(provider=_provider)



import os

from celery import Celery

from .settings import celery_configs


app = Celery("findopendata",
        broker=celery_configs.get("broker", "amqp://"),
        include=[
                "findopendata.ckan_crawler",
                "findopendata.socrata_crawler",
                "findopendata.indexing",
                "findopendata.metadata",
                ])
app.conf.task_default_queue = celery_configs.get("queue")


if __name__ == "__main__":
    app.start()



import numpy as np
import spacy


class _LazySpacyModel:

    def __init__(self, model_name, **kwargs):
        self._model_name = model_name
        self._model = None
        self._model_kwargs = kwargs

    def process(self, text):
        if not self._model:
            self._model = spacy.load(self._model_name, **self._model_kwargs)
        return self._model(text)

    def get_empty_word_vector(self):
        doc = self.process("test")
        return np.zeros(len(doc.vector), dtype=np.float32)


class _SpacyModel(_LazySpacyModel):
    
    def __init__(self, model_name, **kwargs):
        super().__init__(model_name, **kwargs)
        self._model = spacy.load(self._model_name, **self._model_kwargs)




import os
import contextlib
import shutil
import gzip

import simplejson as json
import fastavro

from .base import BlobStorage, Blob


class LocalStorage(BlobStorage):
    """Local storage provider that utilizes the local file system.

    Args:
        root: the root directory, will be created if not exists.
    """

    def __init__(self, root):
        self._root = root
        # Create if not exists.
        if not os.path.exists(self._root):
            os.makedirs(self._root)
    
    def _get_and_check_path(self, blob_name):
        path = os.path.join(self._root, blob_name)
        if not os.path.exists(path):
            raise ValueError("Cannot find blob at path: "+path)
        return path
    
    def _get_path_and_create_dir(self, blob_name):
        path = os.path.join(self._root, blob_name)
        path_dir = os.path.dirname(path)
        if not os.path.exists(path_dir):
            os.makedirs(path_dir)
        return path
    
    def get_object(self, blob_name):
        path = self._get_and_check_path(blob_name)
        with open(path, "r") as f:
            obj = json.load(f)
        return obj

    @contextlib.contextmanager
    def get_file(self, blob_name):
        path = self._get_and_check_path(blob_name)
        try:
            fileobj = open(path, 'rb')
            yield fileobj
        finally:
            fileobj.close()
    
    def put_file(self, fileobj, blob_name):
        path = self._get_path_and_create_dir(blob_name)
        with open(path, 'wb') as f:
            shutil.copyfileobj(fileobj, f)
        size = os.path.getsize(path)
        return Blob(blob_name, size)
    
    def put_object(self, obj, blob_name):
        path = self._get_path_and_create_dir(blob_name)
        with open(path, "w") as f:
            json.dump(obj, f)
        size = os.path.getsize(path)
        return Blob(blob_name, size)
    
    def put_avro(self, schema, records, blob_name, codec='snappy'):
        path = self._get_path_and_create_dir(blob_name)
        with open(path, "wb") as f:
            fastavro.writer(f, schema, records, codec)
        size = os.path.getsize(path)
        return Blob(blob_name, size)
    
    def put_json(self, records, blob_name, gzip_compress=True):
        path = self._get_path_and_create_dir(blob_name)
        newline = "\n"
        if gzip_compress:
            with gzip.open(path, "wt") as f:
                for record in records:
                    f.write(json.dumps(record))
                    f.write(newline)
        else:
            with open(path, "w") as f:
                for record in records:
                    f.write(json.dumps(record))
                    f.write(newline)
        size = os.path.getsize(path)
        return Blob(blob_name, size)



import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="findopendata",
    version="1.0.5",
    author="Eric Zhu",
    author_email="",
    description="A search engine for Open Data.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/findopendata/findopendata",
    keywords="open-data search-engine",
    packages=setuptools.find_packages(exclude=["tests*"]),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=[
        "requests>=2.22.0",
        "google-cloud-storage>=1.17.0",
        "azure-storage-blob>=12.14.1",
        "google-auth>=1.6.3",
        "gcsfs>=0.3.0",
        "celery>=4.3.0",
        "psycopg2-binary>=2.7.5",
        "Django>=2.2.3",
        "rfc6266-content-disposition>=0.0.6",
        "simplejson>=3.16.0",
        "genson>=1.1.0",
        "fastavro>=0.22.3",
        "python-snappy>=0.5.4",
        "python-dateutil>=2.8.0",
        "datasketch>=1.4.10",
        "pyfarmhash>=0.2.2",
        "cchardet>=2.1.4",
        "spacy>=2.1.8",
        "beautifulsoup4>=4.8.0",
        "pyyaml>=5.1.2",
        "gevent>=1.4.0",
    ],
    extras_require={
        "test": [
            "nose2>=0.9.1",
        ]
    },
    scripts=[
        "harvest_datasets.py",
        "generate_metadata.py",
        "sketch_dataset_content.py",
    ],
)



import os
import json
import unittest
import tempfile

from findopendata.socrata import socrata_records
from findopendata.parsers.avro import JSON2AvroRecords
from findopendata.storage.local import LocalStorage

class TestSocrataRecords(unittest.TestCase):

    resource_url = "https://soda.demo.socrata.com/resource/b6kv-3wgw.json"

    def test_demo_api(self):
        records = socrata_records(self.resource_url, app_token=None)
        self.assertTrue(all(isinstance(record, dict) for record in records))
    
    def test_save_avro_records(self):
        records = JSON2AvroRecords(socrata_records(self.resource_url, 
                app_token=None))
        with tempfile.TemporaryDirectory() as root:
            storage = LocalStorage(root)
            json.dump(records.schema, open("schema.json", "w"))
            blob = storage.put_avro(records.schema, records.get(), "test_blob")
            self.assertEqual(blob.name, "test_blob")
            self.assertGreater(blob.size, 0)
            self.assertTrue(os.path.exists(os.path.join(root, "test_blob")))


if __name__ == "__main__":
    unittest.main()




import farmhash
import simplejson as json
import numpy as np
from datasketch import MinHash, HyperLogLogPlusPlus

from .models.word_vector_models import WordVectorModel


def _is_number(x):
    try:
        a = float(x)
    except ValueError:
        return False
    else:
        return True


class ColumnSketch:
    """A Column Sketch contains a summary of a table column. 

    Args:
        column_name: the extracted column name.
        minhash_size: the number of permutations to use for MinHash.
        minhash_seed: the random seed used by MinHash.
        hyperloglog_p: the precision parameter used by HyperLogLog.
        sample_size: the size of sample to be kept.
        enable_word_vector_data: whether to build word embedding vector for 
            data values -- can be 10x more expensive.
    """
    def __init__(self, column_name, 
        minhash_size=256, 
        minhash_seed=43, 
        hyperloglog_p=8,
        sample_size=100,
        enable_word_vector_data=False,
        model=WordVectorModel,
        ):
        self._column_name = column_name
        self._sample = set([])
        self._sample_size = sample_size
        self._count = 0
        self._empty_count = 0
        self._oov_count = 0
        self._numeric_count = 0
        self._minhash = MinHash(num_perm=minhash_size, seed=minhash_seed,
                hashfunc=self._hashfunc32)
        self._hhl = HyperLogLogPlusPlus(p=hyperloglog_p,
                hashfunc=self._hashfunc64) 
        self._enabled_word_vec_data = enable_word_vector_data
        self._model = model
        self._sum_vector = self._model.get_empty_word_vector()

    def _hashfunc32(self, str_value):
        return farmhash.hash32(str_value)
    
    def _hashfunc64(self, str_value):
        return farmhash.hash64(str_value)
    
    @property
    def column_name(self):
        """The extracted column name.
        """
        return self._column_name

    @property
    def sample(self):
        """A sample (non-random) of the data values in the column as a list.
        """
        return list(self._sample)
    
    @property
    def count(self):
        """The total number of data values (i.e. rows) including
        the empty ones.
        """
        return self._count
    
    @property
    def empty_count(self):
        """The number of empty data values.
        """
        return self._empty_count
    
    @property
    def non_empty_count(self):
        """The number of non-empty data values.
        """
        return self._count - self._empty_count
    
    @property
    def out_of_vocabulary_count(self):
        """The number of data values that are non-empty and outside of
        the language model's vocabulary.
        """
        return self._oov_count
    
    @property
    def in_vocabulary_count(self):
        """The number of data values that are non-empty and in
        the language model's vocabulary.
        """
        return self._count - self._empty_count - self._oov_count
    
    @property
    def numeric_count(self):
        """The number of data values that are non-empty and numerical.
        """
        return self._numeric_count
    
    @property
    def is_numeric(self):
        """Whether the column is numeric, based on if at least 50% of rows
        are numeric.
        """
        if self.non_empty_count == 0:
            return False
        return (float(self._numeric_count) / float(self.non_empty_count)) >= 0.5
    
    @property
    def distinct_count(self):
        """The approximate distinct count made by the HyperLogLog.
        """
        if len(self._sample) < self._sample_size:
            return len(self._sample)
        return max(len(self._sample), self._hhl.count())
    
    @property
    def word_vector_column_name(self):
        """The word embedding vector of the column name as a list.
        """
        doc = self._model.process(self.column_name)
        vectors = [token.vector for token in doc if token.has_vector]
        if len(vectors) == 0:
            return None
        return list(float(v) for v in np.sum(vectors, axis=0))
    
    @property
    def word_vector_data(self):
        """The mean word embedding vector of all data values as a list.
        """
        if not self._enabled_word_vec_data:
            return None
        if self.in_vocabulary_count == 0:
            return None
        vector = self._sum_vector / np.float32(self.in_vocabulary_count)
        return list(float(v) for v in vector)
    
    @property
    def minhash(self):
        """The hash values in the MinHash.
        """
        return list(int(v) for v in self._minhash.digest())

    @property
    def seed(self):
        """The random seed used for MinHash.
        """
        return self._minhash.seed
    
    @property
    def hyperloglog(self):
        """The register values of the HyperLogLog counter.
        """
        return list(int(v) for v in self._hhl.digest())

    def update(self, value):
        """Add a data value into the sketch.
        """
        # Update counter.
        self._count += 1
        if not isinstance(value, str):
            value = json.dumps(value, sort_keys=True)
        # Clean the value
        value = value.strip().lower()
        # Skip if the value is empty string.
        if len(value) == 0:
            self._empty_count += 1
            return
        if _is_number(value):
            self._numeric_count += 1
        # Add to sample.
        if len(self._sample) < self._sample_size:
            self._sample.add(value)
        # Update the MinHash sketch.
        self._minhash.update(value)
        # Update the HyperLogLog sketch.
        self._hhl.update(value)
        # Skip word vector extraction if not enabled.
        if not self._enabled_word_vec_data:
            return
        # Update the sum of word embeddings.
        vectors = [token.vector for token in self._model.process(value) 
                if token.has_vector]
        if len(vectors) > 0:
            self._sum_vector += np.sum(vectors, axis=0)
        else:
            self._oov_count += 1




#!/usr/bin/env python
import sys
import argparse
import collections

import psycopg2
from psycopg2.extras import RealDictCursor

from findopendata.settings import db_configs, index_configs
from findopendata.indexing import sketch_package_file


_sql = r"""
WITH updated_times AS (
    SELECT package_file_key as key, max(updated) as updated
    FROM findopendata.column_sketches
    GROUP BY package_file_key
),
package_files AS (
    SELECT f.key, f.blob_name, f.format, f.updated as file_updated,
        u.updated as sketch_updated
    FROM findopendata.package_files as f
    LEFT JOIN updated_times as u
    ON f.key = u.key
)
SELECT key, blob_name, format
FROM package_files
WHERE (sketch_updated IS NULL OR file_updated > sketch_updated)
    AND blob_name IS NOT NULL
"""


_sql_force_update = r"""
SELECT key, blob_name, format
FROM findopendata.package_files
WHERE blob_name IS NOT NULL
"""


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
            description="Creating column sketches of all package files.")
    parser.add_argument("-u", "--force-update", action="store_true")
    args = parser.parse_args(sys.argv[1:])

    print("Creating tasks (force_update = {})...".format(args.force_update))
    package_files = collections.deque([])
    conn = psycopg2.connect(**db_configs)
    cur = conn.cursor(cursor_factory=RealDictCursor)
    cur.execute((_sql_force_update if args.force_update else _sql))
    for row in cur:
        package_files.append(row)
    cur.close()
    conn.close()
    print("Sending {} tasks to workers.".format(len(package_files)))
    for package_file in package_files:
        fmt = package_file["format"].strip().lower()
        sketch_package_file.delay(package_file_key=package_file["key"],
                blob_name=package_file["blob_name"],
                dataset_format=fmt,
                max_records=index_configs["max_records_per_dataset"],
                table_sample_size=index_configs["table_sample_size"],
                minhash_size=index_configs["minhash_size"],
                minhash_seed=index_configs["minhash_seed"],
                hyperloglog_p=index_configs["hyperloglog_p"],
                column_sample_size=index_configs["column_sample_size"],
                enable_word_vector_data=index_configs["enable_word_vector_data"])
    print("Done sending tasks")



import unittest
import io

from findopendata.parsers.csv import csv2json

TEST_CSV_1 = """Column1,Column2,Column3
1,2,3
1,3,4
2,4,5
"""

TEST_CSV_2 = """Title,,
ColumnX,ColumnY,ColumnZ
4,1,3
4,1,4
6,34,123
"""

TEST_CSV_3 = """4,1,3
4,1,4
6,34,123
6,34,123
6,34,123
6,34,123
6,34,123
6,34,123
6,34,123
6,34,123
6,34,123
6,34,123
6,34,123
6,34,123
"""

class TestCSV2JSON(unittest.TestCase):

    def test_csv2json_utf8(self):
        f = io.BytesIO(TEST_CSV_1.encode("utf-8"))
        records = list(csv2json(f))
        self.assertEqual(records[0], {"Column1": "1", "Column2": "2", "Column3": "3"})
        self.assertEqual(records[1], {"Column1": "1", "Column2": "3", "Column3": "4"})
        self.assertEqual(records[2], {"Column1": "2", "Column2": "4", "Column3": "5"})

    def test_csv2json_iso_8859_1(self):
        f = io.BytesIO(TEST_CSV_1.encode("iso-8859-1"))
        records = list(csv2json(f))
        self.assertEqual(records[0], {"Column1": "1", "Column2": "2", "Column3": "3"})
        self.assertEqual(records[1], {"Column1": "1", "Column2": "3", "Column3": "4"})
        self.assertEqual(records[2], {"Column1": "2", "Column2": "4", "Column3": "5"})

    def test_csv2json_header_detect(self):
        f = io.BytesIO(TEST_CSV_2.encode("utf-8"))
        records = list(csv2json(f))
        self.assertEqual(records[0], {"ColumnX": "4", "ColumnY": "1", "ColumnZ": "3"})
        self.assertEqual(records[1], {"ColumnX": "4", "ColumnY": "1", "ColumnZ": "4"})
        self.assertEqual(records[2], {"ColumnX": "6", "ColumnY": "34", "ColumnZ": "123"})

        f = io.BytesIO(TEST_CSV_3.encode("iso-8859-1"))
        records = list(csv2json(f, allow_no_header=True, header_prefix="C"))
        self.assertEqual(records[0], {"C0": "4", "C1": "1", "C2": "3"})


if __name__ == "__main__":
    unittest.main()




import io
import unittest
import json
import gzip

from azure.storage.blob import BlobServiceClient

from findopendata.storage.azure import AzureStorage
from findopendata.parsers.avro import avro2json

test_obj = {
    "name": "First name Last name",
    "email": "example@email.com",
    "age": 42,
    "accounts": [123, 432, 2123],
}

test_file_content = """
h1,h2,h3
a,b,c
e,f,g
1,2,3
"""

test_avro_records = [
    {"h1": "a", "h2": "b", "h3": "c"},
    {"h1": "d", "h2": "e", "h3": "f"},
    {"h1": "1", "h2": "2", "h3": "3"},
]

test_avro_schema = {
    "name": "root",
    "type": "record",
    "fields": [
        {"name": "h1", "type": "string"},
        {"name": "h2", "type": "string"},
        {"name": "h3", "type": "string"},
    ],
}

azurite_connection_str = "DefaultEndpointsProtocol=http;AccountName=devstoreaccount1;AccountKey=Eby8vdM02xNOcqFlqUwJPLlmEtlCDXJ1OUzFT50uSRZ6IFsuFq2UVErCz4I6tq/K1SZFPTOtr/KBHBeksoGMGw==;BlobEndpoint=http://127.0.0.1:10000/devstoreaccount1;"
container_name = "findopendata-test"


class TestAzureStorage(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.service = BlobServiceClient.from_connection_string(azurite_connection_str)
        cls.service.create_container(container_name)

    @classmethod
    def tearDownClass(cls):
        cls.service.delete_container(container_name)

    def setUp(self):
        self.storage = AzureStorage(
            connection_string=azurite_connection_str,
            container_name=container_name,
            block_size=4 * 1024,
        )

    def test_put_and_get_object(self):
        blob = self.storage.put_object(test_obj, "test_object_blob")
        self.assertEqual(blob.name, "test_object_blob")
        obj = self.storage.get_object("test_object_blob")
        self.assertEqual(obj, test_obj)

    def test_put_and_get_file(self):
        fileobj = io.BytesIO(test_file_content.encode("utf-8"))
        blob = self.storage.put_file(fileobj, "test_file_blob")
        self.assertEqual(blob.name, "test_file_blob")
        with self.storage.get_file("test_file_blob") as f:
            data = f.read().decode("utf-8")
            self.assertEqual(data.strip(), test_file_content.strip())

    def test_put_and_get_large_file(self):
        test_large_file_content = test_file_content.encode("utf-8") * 100000
        fileobj = io.BytesIO(test_large_file_content)
        blob = self.storage.put_file(fileobj, "test_large_file_blob")
        self.assertEqual(blob.name, "test_large_file_blob")
        with self.storage.get_file("test_large_file_blob") as f:
            data = f.read()
            self.assertEqual(data.strip(), test_large_file_content.strip())

    def test_put_avro(self):
        blob = self.storage.put_avro(
            test_avro_schema, test_avro_records, "test_avro_blob"
        )
        self.assertEqual(blob.name, "test_avro_blob")
        with self.storage.get_file("test_avro_blob") as f:
            records = avro2json(f)
            for r1, r2 in zip(records, test_avro_records):
                self.assertEqual(dict(r1), r2)

    def test_put_json(self):
        blob = self.storage.put_json(test_avro_records, "test_json_blob")
        self.assertEqual(blob.name, "test_json_blob")

        with self.storage.get_file("test_json_blob") as f:
            f = io.TextIOWrapper(gzip.GzipFile(fileobj=f, mode="rb"))
            records = [json.loads(line) for line in f]
            for r1, r2 in zip(records, test_avro_records):
                self.assertEqual(r1, r2)

    def test_put_avro_large(self):
        test_large_avro_records = test_avro_records * 100000
        blob = self.storage.put_avro(
            test_avro_schema, test_large_avro_records, "test_avro_blob"
        )
        self.assertEqual(blob.name, "test_avro_blob")

        with self.storage.get_file("test_avro_blob") as f:
            records = avro2json(f)
            for r1, r2 in zip(records, test_large_avro_records):
                self.assertEqual(dict(r1), r2)

    def test_put_json_large(self):
        test_large_avro_records = test_avro_records * 100000
        blob = self.storage.put_json(test_large_avro_records, "test_json_blob")
        self.assertEqual(blob.name, "test_json_blob")

        with self.storage.get_file("test_json_blob") as f:
            f = io.TextIOWrapper(gzip.GzipFile(fileobj=f, mode="rb"))
            records = [json.loads(line) for line in f]
            for r1, r2 in zip(records, test_large_avro_records):
                self.assertEqual(r1, r2)


if __name__ == "__main__":
    unittest.main()



import os
import shutil
import uuid

from django.utils.text import get_valid_filename
from contextlib import contextmanager


def get_safe_filename(s, default_filename="unnamed_file"):
    """Returns a sanitized filename from the one given.
    """
    if not s:
        return default_filename
    if len(s) > 255:
        s, ext = os.path.splitext(s)
        ext = ext[:255]
        s = s[:255 - len(ext)]+ ext
    s = s.replace(" ", "_")
    s = get_valid_filename(s).strip(".")
    if len(s) == 0:
        return default_filename
    return s


@contextmanager
def temporary_directory(parent_dir):
    """Creates a temporary directory using with. For example:

        with temporary_directory("/tmp/parent_dir") as dir_name:
            # Do things in the newly created dir_name.
            # ...

        # Now the directory dir_name is removed.
    """
    dir_name = os.path.join(parent_dir, str(uuid.uuid4()))
    os.makedirs(dir_name, exist_ok=True)
    try:
        yield dir_name
    finally:
        shutil.rmtree(dir_name)
